# import pathmagic  # noqa
from data.data_loader import ThreedFrontGroupDataset
from data.variables import GROUPS  # noqa
import data.visualizer as visualizer
import matplotlib.pyplot as plt  # noqa
from sklearn.manifold import TSNE  # noqa
from tqdm import tqdm  # noqa
import numpy as np  # noqa
import importlib
import random  # noqa
import torch
import vedo  # noqa
import time
import json
import sys
import os


PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_parameter_number(net):
    total_num = sum(p.numel() for p in net.parameters())
    trainable_num = sum(p.numel() for p in net.parameters() if p.requires_grad)
    return {'Total': total_num, 'Trainable': trainable_num}


def categorical_kl(p, q):
    return (p * (np.log(p + 1e-6) - np.log(q + 1e-6))).sum()


def main(room_type='bedroom', mode='all_test'):
    ''' mode: forward, forward_9, random, random_9, ground_truth
    '''
    room_types = ['bedroom', 'diningroom', 'library', 'livingroom', 'all']
    TEST_SET = ThreedFrontGroupDataset(room_type, split='train', mode='order', batch_size=1, dataset_len=1000)
    idx = room_types.index(room_type)
    address_list = ['log\\model_completenet_vae_bedroom\\xxxx']
    model_list = ['model_completenet_vae']
    address = os.path.join(PROJECT_DIR, address_list[idx])
    dict_address = os.path.join(address, 'checkpoints/best_model.pth')
    if os.path.join(PROJECT_DIR, 'models') in sys.path:
        sys.path.remove(os.path.join(PROJECT_DIR, 'models'))
    sys.path.append(address)

    MODEL = importlib.import_module(model_list[idx])
    network = MODEL.get_model()
    criterion = MODEL.get_loss()
    criterion = criterion.cuda()
    print(get_parameter_number(network))

    checkpoint = torch.load(dict_address)
    network.load_state_dict(checkpoint['model_state_dict'])
    network = network.cuda().eval()

    with open('./data/complexity_map.json', 'r') as f:
        complexity_map = json.load(f)[room_type]
    f.close()

    if mode == 'forward':
        sample_idx = 100
        bbox, label, bidx, nins, group_idx = TEST_SET.__getitem__(sample_idx)
        bbox = torch.Tensor(bbox).float().cuda()
        label = torch.Tensor(label).long().cuda()
        bidx = torch.Tensor(bidx).long().cuda()
        nins = torch.Tensor(nins).long().cuda()
        group_idx = torch.Tensor([group_idx]).long().cuda()
        pred_output, gt_output, _, _ = network(bbox, label, bidx, nins, group_idx)
        pred_bbox = network.continuous_output(pred_output, type='pred')
        gt_bbox = network.continuous_output(gt_output, type='gt')
        print(f'Group idx: {group_idx}, Pred list: {pred_output["class"].topk(k=1).indices.view(-1)}, GroundTruth: {gt_output["class"]}')
        vis_pred, vis_pred_lab = visualizer.visualize_scene(pred_bbox, pred_output['class'].topk(k=1).indices.view(-1), normalized=True, mode='complete')
        vis_gt, vis_gt_lab = visualizer.visualize_scene(gt_bbox, gt_output['class'], normalized=True, mode='complete')
        vedo.show(vis_gt, vis_gt_lab, axes=2, at=0, N=2)
        vedo.show(vis_pred, vis_pred_lab, axes=2, at=1, interactive=1)

    elif mode == 'forward_9':
        sample_idx = 100
        bbox, label, bidx, nins, group_idx = TEST_SET.__getitem__(sample_idx)
        bbox = torch.Tensor(bbox).float().cuda()
        label = torch.Tensor(label).long().cuda()
        bidx = torch.Tensor(bidx).long().cuda()
        nins = torch.Tensor(nins).long().cuda()
        group_idx = torch.Tensor([group_idx]).long().cuda()
        vis_bbox_list, vis_label_list = [], []
        for i in range(8):
            pred_output, gt_output, _, _ = network(bbox, label, bidx, nins, group_idx)
            pred_bbox = network.continuous_output(pred_output, type='pred')
            vis_pred, vis_pred_lab = visualizer.visualize_scene(pred_bbox, pred_output['class'].topk(k=1).indices.view(-1), normalized=True, mode='complete')
            vis_bbox_list.append(vis_pred)
            vis_label_list.append(vis_pred_lab)
        gt_bbox = network.continuous_output(gt_output, type='gt')
        print(f'Group idx: {group_idx}, Pred list: {pred_output["class"].topk(k=1).indices.view(-1)}, GroundTruth: {gt_output["class"]}')
        vis_gt, vis_gt_lab = visualizer.visualize_scene(gt_bbox, gt_output['class'], normalized=True, mode='complete')
        vedo.show(vis_gt, vis_gt_lab, axes=2, at=0, N=9)
        for i in range(8):
            if i == 7:
                vedo.show(vis_bbox_list[i], vis_label_list[i], axes=2, at=i+1, interactive=1)
            else:
                vedo.show(vis_bbox_list[i], vis_label_list[i], axes=2, at=i+1)

    elif mode == 'random':
        with torch.no_grad():
            group_idx = torch.Tensor([4]).long().cuda()
            nins = torch.Tensor([2]).long().cuda()
            start_sample = time.clock()
            pred_output = network.sample(group_idx[0], nins[0])
            pred_bbox = network.continuous_output(pred_output, type='pred')
            end_sample = time.clock()
            print(f'time: {end_sample-start_sample} type: {pred_output["class"].topk(k=1).indices.view(-1)}')
            model_json = os.path.join(PROJECT_DIR, 'data', 'model_pool.json')
            with open(model_json, 'r') as f:
                model_pool = json.load(f)
            f.close()
            vis_pred, vis_pred_lab = visualizer.visualize_scene(pred_bbox, pred_output['class'].topk(k=1).indices.view(-1), normalized=True, mode='complete', frame=True, model_pool=model_pool)
            vedo.show(vis_pred, vis_pred_lab, axes=2, interactive=1)

    elif mode == 'random_9':
        with torch.no_grad():
            group_idx = torch.Tensor([10]).long().cuda()
            nins = torch.Tensor([2]).long().cuda()
            model_json = os.path.join(PROJECT_DIR, 'data', 'model_pool.json')
            with open(model_json, 'r') as f:
                model_pool = json.load(f)
            f.close()
            vis_bbox_list, vis_label_list = [], []
            number = 9
            for i in range(number):
                pred_output = network.sample(group_idx[0], nins[0])
                pred_bbox = network.continuous_output(pred_output, type='pred')
                vis_pred, vis_pred_lab = visualizer.visualize_scene(pred_bbox, pred_output['class'].topk(k=1).indices.view(-1), normalized=True, mode='complete', frame=True, model_pool=model_pool)
                vis_bbox_list.append(vis_pred)
                vis_label_list.append(vis_pred_lab)
            model_json = os.path.join(PROJECT_DIR, 'data', 'model_pool.json')
            for i in range(number):
                if i == 0:
                    vedo.show(vis_bbox_list[i], vis_label_list[i], axes=2, at=i, N=number)
                if i == number-1:
                    vedo.show(vis_bbox_list[i], vis_label_list[i], axes=2, at=i, interactive=1)
                else:
                    vedo.show(vis_bbox_list[i], vis_label_list[i], axes=2, at=i)

    elif mode == 'ground_truth':
        with torch.no_grad():
            # sample_idx = np.random.randint(0, len(TEST_SET))
            sample_idx = 0
            bbox, label, bidx, nins, group_idx = TEST_SET.__getitem__(sample_idx)
            bbox = torch.Tensor(bbox).float().cuda()
            label = torch.Tensor(label).long().cuda()
            nins = torch.Tensor(nins).long().cuda()
            group_idx = torch.Tensor([group_idx]).long().cuda()
            print(group_idx, list(GROUPS.keys())[group_idx.cpu().numpy()[0]])
            vp = vedo.Plotter(axes=0, interactive=1)
            camera = vp.camera
            camera.SetPosition([0, 20, 20])
            camera.SetViewUp([0, 1, 0])
            vp += vedo.Light(pos=[4, 4, 4], c='white', intensity=0.5)
            vp += vedo.Light(pos=[4, 4, -4], c='white', intensity=0.5)
            vp += vedo.Light(pos=[-4, 4, -4], c='white', intensity=0.5)
            vp += vedo.Light(pos=[-4, 4, 4], c='white', intensity=0.5)
            model_json = os.path.join(PROJECT_DIR, 'data', f'model_pool_{room_type}.json')
            with open(model_json, 'r') as f:
                model_pool = json.load(f)
            f.close()
            vp = visualizer.visualize_synthetic_scene(vp, bbox, label, normalized=True, model_pool=model_pool, numpy=False, vis=False)
            vp.show()


if __name__ == "__main__":
    main(room_type='bedroom', mode='ground_truth')
