# import pathmagic  # noqa
from data.data_loader import ThreedFrontDataset
import data.visualizer as visualizer
import matplotlib.pyplot as plt  # noqa
from sklearn.manifold import TSNE  # noqa
from tqdm import tqdm  # noqa
import numpy as np  # noqa
import importlib
import torch
import time  # noqa
import vedo  # noqa
import sys
import os


PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_parameter_number(net):
    total_num = sum(p.numel() for p in net.parameters())
    trainable_num = sum(p.numel() for p in net.parameters() if p.requires_grad)
    return {'Total': total_num, 'Trainable': trainable_num}


def main():
    DATASET = ThreedFrontDataset(room_type='bedroom', split='train', rot_aug=True)
    mode = 'all'  # sample / gt
    idx = 0
    address_list = ['log\\model_proposenet_bedroom\\xxxx']
    model_list = ['model_proposenet']
    address = os.path.join(PROJECT_DIR, address_list[idx])
    dict_address = os.path.join(address, 'checkpoints/best_model.pth')
    sys.path.remove(os.path.join(PROJECT_DIR, 'models'))
    sys.path.append(address)

    MODEL = importlib.import_module(model_list[idx])
    network = MODEL.get_model()
    criterion = MODEL.get_loss()
    criterion = criterion.cuda()
    print(get_parameter_number(network))

    checkpoint = torch.load(dict_address)
    network.load_state_dict(checkpoint['model_state_dict'])
    network = network.cuda().eval()

    if mode == 'gt':
        group_label = 0
        gt_posrot, pred_label = [], []
        for i in tqdm(range(len(DATASET))):
            group_posrot, group_onehot, group_lab, floor_mask, N, floor_scales = DATASET.__getitem__(i)
            idx = group_lab == group_label
            gt_posrot.append(group_posrot[idx])
            pred_label.append(group_lab[idx])
        gt_posrot = torch.Tensor(np.vstack(gt_posrot)).cuda()
        pred_label = torch.Tensor(np.hstack(pred_label)).cuda()
        vis_pred, vis_pred_lab = visualizer.visualize_scene(gt_posrot, pred_label, normalized=True)
        vedo.show(vis_pred, vis_pred_lab, axes=0, interactive=1)

    elif mode == 'sample':
        with torch.no_grad():
            group_lab = torch.Tensor([0]).long().cuda()  # change the group label
            gt_posrot, pred_label = network.sample_single_group(group_lab, 500)
            vis_pred, vis_pred_lab = visualizer.visualize_scene(gt_posrot, pred_label, normalized=True)
            vedo.show(vis_pred, vis_pred_lab, axes=2, interactive=1)


if __name__ == "__main__":
    main()
