import pathmagic  # noqa
from data.data_loader import ThreedFrontDataset
from torch.utils.data import DataLoader
from model import ProposeNet
import torch.nn.functional as F  # noqa
import torch.nn as nn
import numpy as np
import torch
import time
import os


os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'


class get_model(nn.Module):
    def __init__(self, ndim=4, nclass=12):
        super(get_model, self).__init__()
        self.ndim = ndim
        self.nclass = nclass
        self.cmaf = ProposeNet(ndim=4, ndimcon=self.nclass, nlayer=8)

    def forward(self, group_posrot, group_onehot, N):
        B = group_posrot.shape[0]
        posrot = self.extract_valid_input(group_posrot, N, B)
        posrot = self.dequantify_rot(posrot, scale=0.1)
        onehot = self.extract_valid_input(group_onehot, N, B)
        logprob = self.cmaf(posrot, onehot)
        return logprob

    def sample_single_group(self, group_lab, N):
        group_lab = group_lab.repeat(N).view(N, -1).transpose(0, 1).contiguous().view(-1).contiguous()
        onehot = torch.eye(self.nclass)[group_lab].cuda()
        sample_posrot = self.cmaf.sample(onehot)
        sample_posrot = self.quantify_rot(sample_posrot)
        return sample_posrot, group_lab

    def extract_valid_input(self, input, N, B):
        valid_data = []
        for i in range(N.shape[0]):
            valid_data.append(input[i, :N[i]])
        return torch.cat(valid_data, dim=0)

    def extract_valid_floor_feat(self, floor_feat, N, B):
        feat = []
        for i in range(B):
            feat.append(floor_feat[i].view(1, -1).repeat((N[i], 1)).contiguous())
        return torch.cat(feat, dim=0)

    def dequantify_rot(self, posrot, scale):
        N = posrot.shape[0]
        posrot[:, 3] = posrot[:, 3] + torch.randn(N).cuda() * scale
        return posrot

    def quantify_rot(self, posrot, bin=2):
        rot = torch.clamp(posrot[:, 3], min=-1, max=1)
        rot = torch.round(rot * bin + bin)
        rot = (rot - bin) / bin
        posrot[:, 3] = rot
        return posrot


class get_loss(nn.Module):
    def __init__(self):
        super(get_loss, self).__init__()

    def forward(self, logprob):
        loss_cmaf = self.get_cmaf_loss(logprob)
        return loss_cmaf

    def get_cmaf_loss(self, logprob):
        return -torch.mean(logprob)


if __name__ == '__main__':
    from tqdm import tqdm
    TRAINING_SET = ThreedFrontDataset()
    trainDataLoader = DataLoader(TRAINING_SET, batch_size=16, shuffle=True, num_workers=0, pin_memory=True,
                                 drop_last=True, worker_init_fn=lambda x: np.random.seed(x+int(time.time())))
    network = get_model().cuda()
    criterion = get_loss().cuda()
    bar = tqdm(enumerate(trainDataLoader), total=len(trainDataLoader), smoothing=0.9)
    for i, data in bar:
        group_posrot, group_onehot, group_lab, floor_mask, N, _ = data
        group_posrot = group_posrot.float().cuda()
        group_onehot = group_onehot.float().cuda()
        group_lab = group_lab.long().cuda()
        floor_mask = floor_mask.float().cuda()
        N = N.long().cuda()

        logprob = network(group_posrot, group_onehot, N)
        loss = criterion(logprob)
        print(loss)
        break
