import pathmagic  # noqa
from torch.distributions import MultivariateNormal
import matplotlib.pyplot as plt  # noqa
from sklearn.manifold import TSNE  # noqa
from data.variables import GROUP_COLOR
import data.visualizer as visualizer
from tqdm import tqdm  # noqa
import numpy as np  # noqa
import torch.nn as nn  # noqa
import random  # noqa
import torch
import vedo  # noqa
import math
import roma
import time  # noqa
import json
import copy
import cv2
import os


PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))
EPS = 1e-4


class OptimizationModel():
    def __init__(self):
        self.sigma = 0.1
        self.alpha = 0.25
        self.beta = 0.01
        self.prior = MultivariateNormal(torch.zeros(2).cuda(), self.sigma**2*torch.eye(2).cuda())

    def run(self, posrot_selected, cpts_ori_list, bbox_ori_list, floor_cpts, th, group_lab, max_iter=10):
        posrot_selected_copy = copy.deepcopy(posrot_selected)
        n_bbox = posrot_selected_copy.shape[0]
        cpts_init_list = self.update_cpts(posrot_selected, cpts_ori_list, n_bbox)
        scores_curr, _, _, _ = self.calculate_score(posrot_selected_copy, cpts_init_list, floor_cpts, group_lab, th, n_bbox)
        for i in range(max_iter):
            posrot_updated = self.update_pos(posrot_selected_copy)
            cpts_updated_list = self.update_cpts(posrot_updated, cpts_ori_list, n_bbox)
            scores_updated, _, _, _ = self.calculate_score(posrot_updated, cpts_updated_list, floor_cpts, group_lab, th, n_bbox)
            idx = scores_updated > scores_curr
            posrot_selected_copy[idx] = posrot_updated.clone()[idx]
            scores_curr[idx] = scores_updated[idx]
            if scores_curr.sum() == posrot_updated.shape[0] * 0.90:
                break
        bbox_opt_list = self.update_bbox(posrot_selected_copy, bbox_ori_list, n_bbox)
        return bbox_opt_list

    def run_time_statitics(self, posrot_selected, cpts_ori_list, bbox_ori_list, floor_cpts, th, group_lab, floor_mesh, max_iter=10):
        posrot_selected_copy = copy.deepcopy(posrot_selected)
        n_bbox = posrot_selected_copy.shape[0]
        cpts_init_list = self.update_cpts(posrot_selected, cpts_ori_list, n_bbox)
        scores_curr, _, _, _ = self.calculate_score(posrot_selected_copy, cpts_init_list, floor_cpts, group_lab, th, n_bbox)
        for i in range(max_iter):
            posrot_updated = self.update_pos(posrot_selected_copy)
            cpts_updated_list = self.update_cpts(posrot_updated, cpts_ori_list, n_bbox)
            scores_updated, _, _, _ = self.calculate_score(posrot_updated, cpts_updated_list, floor_cpts, group_lab, th, n_bbox)
            idx = scores_updated > scores_curr
            posrot_selected_copy[idx] = posrot_updated.clone()[idx]
            scores_curr[idx] = scores_updated[idx]
            if scores_curr.sum() == posrot_updated.shape[0]:
                break
        bbox_opt_list = self.update_bbox(posrot_selected_copy, bbox_ori_list, n_bbox)
        cpts_opt_list = self.update_cpts(posrot_selected_copy, cpts_ori_list, n_bbox)
        return bbox_opt_list, cpts_opt_list

    def update_pos(self, posrot_selected):
        posrot_updated = posrot_selected.clone()
        delta_x, delta_z = self.delta_pos(posrot_selected.shape[0])
        posrot_updated[:, 0] = posrot_updated[:, 0] + delta_x
        posrot_updated[:, 2] = posrot_updated[:, 2] + delta_z
        return posrot_updated

    def update_cpts(self, posrot_updated, cpts_ori_list, n_bbox):
        cpts_updated_list = []
        for i in range(n_bbox):
            cpts_updated_list.append(self.transforme_cpts(posrot_updated[i, 0], posrot_updated[i, 2], posrot_updated[i, 3], cpts_ori_list[i]))
        return cpts_updated_list

    def update_bbox(self, posrot_updated, bbox_ori_list, n_bbox):
        bbox_updated_list = []
        for i in range(n_bbox):
            bbox_updated_list.append(self.transform_group(posrot_updated[i, 0], posrot_updated[i, 2], posrot_updated[i, 3], bbox_ori_list[i]))
        return bbox_updated_list

    def delta_pos(self, n_bbox):
        delta = self.prior.sample((n_bbox,)) * self.alpha
        return delta[:, 0], delta[:, 1]

    def delta_rot(self):
        pass

    def calculate_score(self, posrot_updated, cpts_updated_list, floor_cpts, group_lab, th, n_bbox):
        npts_out_floor, cpts_outzone_list = self.caluclate_outzone_count(cpts_updated_list, floor_cpts, th, n_bbox)
        npts_collisoin, cpts_collision_list = self.calculate_collision_count(cpts_updated_list, group_lab, th, n_bbox)
        npts_back, cpts_back_list = self.calcualte_back_count(posrot_updated, cpts_updated_list, floor_cpts, group_lab, th, n_bbox)
        # print(npts_out_floor, npts_collisoin, npts_back)
        npts = []
        for i in range(len(cpts_updated_list)):
            npts.append(cpts_updated_list[i].shape[0])
        npts = torch.Tensor(npts).cuda().long()
        # print(npts_out_floor, npts_collisoin, npts_back)
        scores = (npts - npts_out_floor - npts_collisoin - npts_back) / npts
        # scores = (npts - npts_out_floor - npts_back) / npts
        return scores, cpts_outzone_list, cpts_collision_list, cpts_back_list

    def calculate_collision_count(self, cpts_updated_list, group_lab, th, n_bbox):
        scores = torch.zeros(n_bbox, n_bbox).cuda()
        pts = []
        for i in range(n_bbox):
            for j in range(n_bbox):
                if i == j:
                    scores[i, j] = 0
                else:
                    if group_lab[i] == 3 or group_lab[j] == 3:
                        scores[i, j] = 0
                    else:
                        scores[i, j], p = self.cpts_intersection(cpts_updated_list[i], cpts_updated_list[j], th, reverse=False)
                        pts.append(p)
        return scores.mean(dim=1), pts

    def calcualte_back_count(self, posrot_updated, cpts_updated_list, floor_cpts, group_lab, th, n_bbox):
        scores = []
        pts = []
        for i in range(n_bbox):
            if group_lab[i] != 3:
                s, p = self.cpts_at_back(cpts_updated_list[i], floor_cpts, posrot_updated[i, 3], th)
                scores.append(s)
                pts.append(p)
            else:
                scores.append(0)
        return torch.Tensor(scores).cuda(), pts

    def caluclate_outzone_count(self, cpts_updated_list, floor_cpts, th, n_bbox):
        count = []
        pts = []
        for i in range(n_bbox):
            c, p = self.cpts_intersection(cpts_updated_list[i], floor_cpts, th, reverse=True)
            count.append(c)
            pts.append(p)
        return torch.Tensor(count).cuda(), pts

    def change_or_not(self, scores_curr, scores_updated):
        dice = torch.rand(1).cuda()
        alpha = torch.exp(self.beta*(scores_curr - scores_updated))
        if dice < alpha:
            return True, alpha
        else:
            return False, alpha

    def cpts_intersection(self, cpts1, cpts2, th, reverse=True):
        N = cpts1.shape[0]
        M = cpts2.shape[0]
        dist = -2 * torch.matmul(cpts1, cpts2.permute(1, 0))
        dist += torch.sum(cpts1 ** 2, -1).view(N, 1)
        dist += torch.sum(cpts2 ** 2, -1).view(1, M)
        dist = torch.sqrt(torch.clamp(dist, min=EPS))
        npts_idx = torch.sum(dist <= th+EPS, dim=-1) > 0
        npts_intersect = torch.sum(npts_idx, dim=-1)
        if reverse:
            return N - npts_intersect, cpts1[torch.logical_not(npts_idx)]
        else:
            return npts_intersect, cpts1[npts_idx]

    def cpts_at_back(self, cpts, floor_cpts, rot, th):
        cano_rot = torch.Tensor([0, 90, 180, 270]).cuda()
        maxs = torch.max(cpts, dim=0).values
        mins = torch.min(cpts, dim=0).values
        rot_selection = torch.argmin(torch.abs(cano_rot - (rot * 180 + 180) % 360), dim=-1)
        if rot_selection == 0:
            line_idx = floor_cpts[:, 2] < mins[2]  # -z
            x_minus_idx = floor_cpts[:, 0] > mins[0]
            x_plus_idx = floor_cpts[:, 0] < maxs[0]
            # n_width = torch.abs(maxs[0] - mins[0])
            n = line_idx & x_minus_idx & x_plus_idx
        elif rot_selection == 1:
            line_idx = floor_cpts[:, 0] < mins[0]  # -x
            z_minus_idx = floor_cpts[:, 2] > mins[2]
            z_plus_idx = floor_cpts[:, 2] < maxs[2]
            # n_width = torch.abs(maxs[2] - mins[2])
            n = line_idx & z_minus_idx & z_plus_idx
        elif rot_selection == 2:
            line_idx = floor_cpts[:, 2] > maxs[2]  # z
            x_minus_idx = floor_cpts[:, 0] > mins[0]
            x_plus_idx = floor_cpts[:, 0] < maxs[0]
            # n_width = torch.abs(maxs[0] - mins[0])
            n = line_idx & x_minus_idx & x_plus_idx
        elif rot_selection == 3:
            line_idx = floor_cpts[:, 0] > maxs[0]  # -x
            z_minus_idx = floor_cpts[:, 2] > mins[2]
            z_plus_idx = floor_cpts[:, 2] < maxs[2]
            # n_width = torch.abs(maxs[2] - mins[2])
            n = line_idx & z_minus_idx & z_plus_idx
        pts_back = floor_cpts[n]
        score = pts_back.shape[0]
        return score, pts_back

    def transform_group(self, pos_x, pos_z, rot, group_bbox):
        rot = rot * 180 + 180
        group_rot = (group_bbox[:, 3] * 180 + 180) + rot
        group_bbox_adapted = group_bbox.clone()
        pts = group_bbox[:, 0:3]
        R = roma.rotvec_to_rotmat((rot / 360.) * 2 * math.pi * torch.Tensor([0, 1, 0]).cuda())
        pts = torch.matmul(R, pts.transpose(0, 1)).transpose(0, 1)
        pts[:, 0] = pts[:, 0] + pos_x
        pts[:, 2] = pts[:, 2] + pos_z
        group_bbox_adapted[:, :3] = pts
        group_bbox_adapted[:, 3] = ((group_rot % 360) - 180.0) / 180.0
        return group_bbox_adapted

    def transforme_cpts(self, pos_x, pos_z, rot, check_pts):
        rot = rot * 180 + 180
        R = roma.rotvec_to_rotmat((rot / 360.) * 2 * math.pi * torch.Tensor([0, 1, 0]).cuda())
        check_pts_adapted = torch.matmul(R, check_pts.transpose(0, 1)).transpose(0, 1)
        check_pts_adapted[:, 0] = check_pts_adapted[:, 0] + pos_x * 4
        check_pts_adapted[:, 2] = check_pts_adapted[:, 2] + pos_z * 4
        return check_pts_adapted

    def vis_cpts(self, cpts_list, group_lab, cpts_outzone_list=None, cpts_collision_list=None, cpts_back_list=None):
        vis = []
        for i in range(len(cpts_list)):
            cpts = cpts_list[i].cpu().numpy()
            vis.append(vedo.Points(cpts, c=GROUP_COLOR[group_lab[i].cpu().numpy().astype('int16')], r=7))
            # vis.append(vedo.Points(cpts, c='black', r=7))
        if cpts_outzone_list is not None:
            for i in range(len(cpts_outzone_list)):
                cpts_outzone = cpts_outzone_list[i].cpu().numpy()
                vis.append(vedo.Points(cpts_outzone, c='red', r=7))
        if cpts_collision_list is not None:
            for i in range(len(cpts_collision_list)):
                cpts_collision = cpts_collision_list[i].cpu().numpy()
                vis.append(vedo.Points(cpts_collision, c='green', r=7))
        if cpts_back_list is not None:
            for i in range(len(cpts_back_list)):
                cpts_back = cpts_back_list[i].cpu().numpy()
                vis.append(vedo.Points(cpts_back, c='cyan', r=7))
        return vis

    def vis_bbox(self, bbox_list):
        vis = []
        for i in range(len(bbox_list)):
            vis_bbox, vis_frame, _ = visualizer.visualize_scene(bbox_list[i][:, 0:7], bbox_list[i][:, 7], normalized=True, mode='complete', only_model=False)
            vis += vis_frame
        return vis

    def vis_floor_cpts(self, floor_cpts):
        vis = vedo.Points(floor_cpts.cpu().numpy(), r=3)
        return [vis]


class CheckPointUtils():
    def __init__(self, n_proposal):
        self.curr_cpts = None
        self.cano_rot = torch.Tensor([0, 90, 180, 270]).cuda().repeat(n_proposal).view(n_proposal, 4).contiguous()

    def get_best_idx(self, bbox_trans, cpts_trans, rot, floor_cpts, th, n_proposal, label):
        if label == 3:
            score = torch.abs(bbox_trans[:, 0, 0]) + torch.abs(bbox_trans[:, 0, 2])
            return torch.argmin(score).view(1).long().cuda()
            # return torch.randint(0, n_proposal).view(1).long().cuda()
        npts = cpts_trans.shape[1]
        # number of cpts out of floor
        npts_out_floor = self.cpts_intersection(cpts_trans, floor_cpts, th, n_proposal, floor=True)
        # number of cpts collided by other cpts
        npts_collision = self.cpts_intersection(cpts_trans, self.curr_cpts, th, n_proposal, floor=False)
        # number of cpts at the back of the group
        npts_back = self.cpts_at_back(cpts_trans, floor_cpts, rot, n_proposal)
        best_idx = (npts - npts_out_floor - npts_collision - npts_back).topk(k=1).indices
        # print(npts - npts_out_floor - npts_collision, npts_back, best_idx)
        if self.curr_cpts is None:
            self.curr_cpts = cpts_trans[best_idx].view(-1, 3).contiguous()
        else:
            self.curr_cpts = torch.cat([self.curr_cpts, cpts_trans[best_idx].view(-1, 3).contiguous()], dim=0)
        # print(bbox_trans[best_idx].shape, cpts_trans[best_idx].shape)
        return best_idx

    def reset(self):
        self.curr_cpts = None

    def cpts_at_back(self, cpts_trans, floor_cpts, rot, n_proposal):
        maxs = torch.max(cpts_trans, dim=1).values
        mins = torch.min(cpts_trans, dim=1).values
        # scales = maxs - mins
        rot_selection = torch.argmin(torch.abs(self.cano_rot - ((rot * 180 + 180) % 360).view(n_proposal, 1).contiguous()), dim=-1)
        # print(maxs, mins, scales, rot_selection)
        npts_back = []
        for i in range(n_proposal):
            if rot_selection[i] == 0:
                line_idx = floor_cpts[:, 2] < mins[i, 2]  # -z
                x_minus_idx = floor_cpts[:, 0] > mins[i, 0]
                x_plus_idx = floor_cpts[:, 0] < maxs[i, 0]
                n = line_idx & x_minus_idx & x_plus_idx
            elif rot_selection[i] == 1:
                line_idx = floor_cpts[:, 0] < mins[i, 0]  # -x
                z_minus_idx = floor_cpts[:, 2] > mins[i, 2]
                z_plus_idx = floor_cpts[:, 2] < maxs[i, 2]
                n = line_idx & z_minus_idx & z_plus_idx
            elif rot_selection[i] == 2:
                line_idx = floor_cpts[:, 2] > maxs[i, 2]  # z
                x_minus_idx = floor_cpts[:, 0] > mins[i, 0]
                x_plus_idx = floor_cpts[:, 0] < maxs[i, 0]
                n = line_idx & x_minus_idx & x_plus_idx
            elif rot_selection[i] == 3:
                line_idx = floor_cpts[:, 0] > maxs[i, 0]  # -x
                z_minus_idx = floor_cpts[:, 2] > mins[i, 2]
                z_plus_idx = floor_cpts[:, 2] < maxs[i, 2]
                n = line_idx & z_minus_idx & z_plus_idx
            npts_back.append(n.sum().view(-1).contiguous())
        return torch.cat(npts_back, dim=0)

    def cpts_intersection(self, cpts1, cpts2, th, B, floor=False):
        ''' cpts1: B, N, 3
            cpts2: floor or curr_pts
        '''
        if cpts2 is None:
            return 0
        cpts2_clone = cpts2.view(1, -1, 3).contiguous().repeat((B, 1, 1))
        N = cpts1.shape[1]
        M = cpts2_clone.shape[1]
        dist = -2 * torch.bmm(cpts1, cpts2_clone.transpose(2, 1))
        dist += torch.sum(cpts1 ** 2, -1).view(B, N, 1)
        dist += torch.sum(cpts2_clone ** 2, -1).view(B, 1, M)
        dist = torch.sqrt(torch.clamp(dist, min=EPS))
        if floor:
            npts = N - torch.sum(torch.sum(dist <= th, dim=-1) > 0, dim=-1)
        else:
            npts = torch.sum(torch.sum(dist <= th, dim=-1) > 0, dim=-1)
        return npts

    def transform_batch_cpts(self, pos_x, pos_z, rot, cpts, n_proposal):
        batch_cpts = cpts.repeat((n_proposal, 1, 1))
        batch_rot = (rot * 180 + 180).view(n_proposal, 1).contiguous()
        R = roma.rotvec_to_rotmat((batch_rot / 360.) * 2 * math.pi * torch.Tensor([0, 1, 0]).cuda())
        cpts_adapted = torch.matmul(R, batch_cpts.transpose(1, 2)).transpose(2, 1).contiguous()
        cpts_adapted[:, :, 0] = cpts_adapted[:, :, 0] + pos_x.view(-1, 1).contiguous() * 4
        cpts_adapted[:, :, 2] = cpts_adapted[:, :, 2] + pos_z.view(-1, 1).contiguous() * 4
        return cpts_adapted.contiguous()

    def transform_batch_group(self, pos_x, pos_z, rot, bbox, n_proposal):
        batch_bbox = bbox.repeat((n_proposal, 1, 1))
        batch_rot = (rot * 180 + 180).view(n_proposal, 1).contiguous()
        batch_group_rot = (batch_bbox[:, :, 3] * 180 + 180) + batch_rot
        batch_bbox_adpated = batch_bbox.clone()
        pts = batch_bbox[:, :, 0:3]
        R = roma.rotvec_to_rotmat((batch_rot / 360.) * 2 * math.pi * torch.Tensor([0, 1, 0]).cuda())
        pts = torch.bmm(R, pts.transpose(1, 2)).transpose(2, 1)
        pts[:, :, 0] = pts[:, :, 0] + pos_x.view(-1, 1).contiguous()
        pts[:, :, 2] = pts[:, :, 2] + pos_z.view(-1, 1).contiguous()
        batch_bbox_adpated[:, :, 0:3] = pts
        batch_bbox_adpated[:, :, 3] = ((batch_group_rot % 360) - 180.0) / 180.0
        return batch_bbox_adpated.contiguous()


class VisualizerUtils():
    def __init__(self, room_type='bedroom'):
        self.floor_mesh = None
        self.vis_floor_cpts = None
        self.vis_prop, self.vis_prop_lab = None, None

        self.model_json = os.path.join(PROJECT_DIR, 'data', f'model_pool_{room_type}.json')
        with open(self.model_json, 'r') as f:
            self.model_pool = json.load(f)
        f.close()

        self.bbox_ori = []
        self.vis_cpts_ori = []

        self.bbox_trans = []
        self.vis_cpts_trans = []

        self.bbox_best = []
        self.vis_cpts_best = []

        self.bbox_opt = []

    def add_floor(self, floor_mesh, floor_cpts):
        self.floor_mesh = floor_mesh
        self.vis_floor_cpts = self.visualize_check_points(floor_cpts, r=3)

    def add_proposal(self, pred_posrot, pred_label):
        pred_posrot_clone = pred_posrot.clone()
        pred_label_clone = pred_label.clone()
        self.vis_prop, self.vis_prop_lab = visualizer.visualize_scene(pred_posrot_clone, pred_label_clone, normalized=True, mode='proposal')

    def add_ori_group(self, group_bbox, group_cpts, group_lab):
        self.bbox_ori.append(group_bbox)
        self.vis_cpts_ori.append(self.visualize_check_points(group_cpts, group_lab, r=7))

    def add_trans_group(self, bbox_trans, cpts_trans, group_lab):
        self.bbox_trans.append(bbox_trans.view(-1, 9).contiguous())
        self.vis_cpts_trans.append(self.visualize_check_points(cpts_trans.view(-1, 3).contiguous(), group_lab, r=7))

    def add_best_group(self, bbox_best, cpts_best, group_lab):
        self.bbox_best.append(bbox_best.view(-1, 9).contiguous())
        self.vis_cpts_best.append(self.visualize_check_points(cpts_best.view(-1, 3).contiguous(), group_lab, r=7))

    def add_opt_group(self, bbox_opt):
        self.bbox_opt.append(bbox_opt.view(-1, 9).contiguous())

    def visualize_check_points(self, check_pts, group_lab=None, idx_collision=None, r=5):
        check_pts = check_pts.detach().cpu().numpy()
        if idx_collision is not None:
            idx_collision = idx_collision.cpu().numpy()
        if group_lab is not None:
            label = group_lab.cpu().numpy().astype('int16')
            if idx_collision is not None:
                pts = vedo.Points(check_pts[idx_collision], GROUP_COLOR[label], r=r)
            else:
                pts = vedo.Points(check_pts, GROUP_COLOR[label], r=r)
        else:
            pts = vedo.Points(check_pts, r=r)
        return pts

    def reset(self):
        self.floor_mesh = None
        self.vis_floor_cpts = None
        self.vis_prop, self.vis_prop_lab = None, None

        self.bbox_ori = []
        self.vis_cpts_ori = []

        self.bbox_trans = []
        self.vis_cpts_trans = []

        self.bbox_best = []
        self.vis_cpts_best = []

        self.bbox_opt = []

    def save_synthetic_image(self, save_dir, save_name):
        vedo.settings.useParallelProjection = 1
        vp = vedo.Plotter(axes=0, offscreen=True, interactive=1)
        camera = vp.camera
        camera.SetPosition([0, 5, 0])
        camera.SetViewUp([0, -1, 0])
        for i in range(len(self.bbox_best)):
            vp = visualizer.visualize_synthetic_scene(vp, self.bbox_best[i][:, :7], self.bbox_best[i][:, 7], normalized=True,  model_pool=self.model_pool)
        vp += self.floor_mesh
        vp.show()
        vedo.screenshot(os.path.join(save_dir, save_name))
        vp.close()
        img = cv2.imread(os.path.join(save_dir, save_name))
        img = cv2.resize(img, (256, 256))
        cv2.imwrite(os.path.join(save_dir, save_name), img)
        self.reset()

    def save_synthetic_image_opt(self, save_dir, save_name):
        vedo.settings.useParallelProjection = 1
        vp = vedo.Plotter(axes=0, offscreen=True, interactive=1)
        camera = vp.camera
        camera.SetPosition([0, 5, 0])
        camera.SetViewUp([0, -1, 0])
        for i in range(len(self.bbox_best)):
            vp = visualizer.visualize_synthetic_scene(vp, self.bbox_opt[i][:, :7], self.bbox_opt[i][:, 7], normalized=True,  model_pool=self.model_pool)
        vp += self.floor_mesh
        vp.show()
        vedo.screenshot(os.path.join(save_dir, save_name))
        vp.close()
        img = cv2.imread(os.path.join(save_dir, save_name))
        img = cv2.resize(img, (256, 256))
        cv2.imwrite(os.path.join(save_dir, save_name), img)
        self.reset()

    def show_opt(self):
        vis_bbox_best = []
        for i in range(len(self.bbox_best)):
            vis_best, vis_frame_best = visualizer.visualize_scene(self.bbox_best[i][:, :7], self.bbox_best[i][:, 7], normalized=True, mode='complete', only_model=True, model_pool=self.model_pool, frame=False)
            vis_bbox_best += vis_best
        vis_bbox_opt = []
        for i in range(len(self.bbox_opt)):
            vis_opt, lab_opt = visualizer.visualize_scene(self.bbox_opt[i][:, :7], self.bbox_opt[i][:, 7], normalized=True, mode='complete', model_pool=self.model_pool, only_model=True, frame=False)
            vis_bbox_opt += vis_opt
        vedo.show(vis_bbox_best + [self.floor_mesh], axes=0, at=0, N=2)
        vedo.show(vis_bbox_opt + [self.floor_mesh], axes=0, at=1, interactive=1)

    def show(self):
        # the original group generated in canonical coord system
        self.bbox_ori = torch.cat(self.bbox_ori, dim=0)
        vis_bbox_ori, vis_ori_label = visualizer.visualize_scene(self.bbox_ori[:, :7], self.bbox_ori[:, 7], normalized=True, mode='complete')
        # the transformed group by area posrot
        self.bbox_trans = torch.cat(self.bbox_trans, dim=0)
        vis_bbox_trans, vis_label_trans = visualizer.visualize_scene(self.bbox_trans[:, :7], self.bbox_trans[:, 7], normalized=True, mode='complete', frame=False)
        # the best selected group bbox
        vis_bbox_best, vis_bbox_frame_best, vis_label_best = [], [], []
        for i in range(len(self.bbox_best)):
            vis_best, vis_frame_best, lab_best = visualizer.visualize_scene(self.bbox_best[i][:, :7], self.bbox_best[i][:, 7], normalized=True, mode='complete', only_model=False, model_pool=self.model_pool, frame=False)
            vis_bbox_best.append(vis_best)
            vis_bbox_frame_best.append(vis_frame_best)
            vis_label_best.append(lab_best)
        # the optimized group bbox
        vis_bbox_opt, vis_label_opt = [], []
        for i in range(len(self.bbox_opt)):
            vis_opt, lab_opt = visualizer.visualize_scene(self.bbox_opt[i][:, :7], self.bbox_opt[i][:, 7], normalized=True, mode='complete', model_pool=self.model_pool, only_model=True, frame=False)
            vis_bbox_opt.append(vis_opt)
            vis_label_opt.append(lab_opt)
        vedo.show([self.vis_prop, self.floor_mesh, self.vis_floor_cpts], self.vis_prop_lab, axes=2, at=0, N=5)
        vedo.show([vis_bbox_best, vis_bbox_frame_best, self.floor_mesh, self.vis_floor_cpts, self.vis_cpts_best], vis_label_best, axes=2, at=1)
        vedo.show([vis_bbox_best, vis_bbox_frame_best, self.floor_mesh], vis_label_best, axes=2, at=2)
        vedo.show([vis_bbox_best, self.floor_mesh], None, axes=0, at=3)
        vedo.show([vis_bbox_opt, self.floor_mesh], None, axes=0, at=4, interactive=1)
